from typing import TYPE_CHECKING

from saefarer.protocols import EncodingOutput

if TYPE_CHECKING:
    from transformers import (
        BertTokenizer,
        BertTokenizerFast,
        RobertaTokenizer,
        RobertaTokenizerFast,
    )


class HuggingFaceBertTokenizerAdapter:
    def __init__(
        self,
        tokenizer: "BertTokenizer | BertTokenizerFast",
    ):
        self.tokenizer = tokenizer
        self.pad_token_id: int = tokenizer.pad_token_id  # type: ignore

    def encode(
        self,
        text: str,
        max_length: int,
    ):
        output = self.tokenizer(
            text,
            padding="max_length",
            max_length=max_length,
            truncation=True,
        )

        token_ids: list[int] = output["input_ids"]  # type: ignore
        attention_mask: list[int] = output["attention_mask"]  # type: ignore

        return EncodingOutput(token_ids=token_ids, attention_mask=attention_mask)

    def decode(self, token_ids: list[int]) -> str:
        tokens = []

        for token in self.tokenizer.convert_ids_to_tokens(token_ids):
            if token.startswith("##"):
                tokens.append(token[2:])
            else:
                tokens.append(" " + token)

        return "".join(tokens)


class HuggingFaceRobertaTokenizerAdapter:
    def __init__(
        self,
        tokenizer: "RobertaTokenizer | RobertaTokenizerFast",
    ):
        self.tokenizer = tokenizer
        self.pad_token_id: int = tokenizer.pad_token_id  # type: ignore

    def encode(
        self,
        text: str,
        max_length: int,
    ):
        output = self.tokenizer(
            text,
            padding="max_length",
            max_length=max_length,
            truncation=True,
        )

        token_ids: list[int] = output["input_ids"]  # type: ignore
        attention_mask: list[int] = output["attention_mask"]  # type: ignore

        return EncodingOutput(token_ids=token_ids, attention_mask=attention_mask)

    def decode(self, token_ids: list[int]) -> str:
        return self.tokenizer.decode(token_ids)



import json
from dataclasses import asdict
from importlib import import_module
from pathlib import Path
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from os import PathLike

    from saefarer.training.config import TrainingConfig


class LogData(TypedDict):
    elapsed_seconds: float
    n_training_batches: int
    n_training_tokens: int
    loss: float
    mse_loss: float
    aux_loss: float
    n_dead_features: int
    mean_n_batches_since_fired: float
    max_n_batches_since_fired: int


class Logger:
    def __init__(self, cfg: "TrainingConfig", log_dir: "str | PathLike"):
        self.cfg = cfg
        self.log_dir = Path(log_dir)

    def write(self, data: LogData):
        pass

    def close(self):
        pass


class WAndBLogger(Logger):
    def __init__(self, cfg: "TrainingConfig", log_dir: "str | PathLike"):
        super().__init__(cfg, log_dir)

        self.wandb = import_module("wandb")

        self.wandb.init(
            config=asdict(cfg),
            project=cfg.wandb_project,
            group=cfg.wandb_group,
            name=cfg.wandb_name,
            notes=cfg.wandb_notes,
            dir=self.log_dir,
        )

    def write(self, data: LogData):
        self.wandb.log(data=data, step=data["n_training_batches"])

    def close(self):
        self.wandb.finish()


class TensorboardLogger(Logger):
    def __init__(self, cfg: "TrainingConfig", log_dir: "str | PathLike"):
        super().__init__(cfg, log_dir)

        from torch.utils.tensorboard.writer import SummaryWriter

        self.writer = SummaryWriter(self.log_dir)

    def write(self, data: LogData):
        for [key, value] in data.items():
            self.writer.add_scalar(key, value, global_step=data["n_training_batches"])
        self.writer.flush()

    def close(self):
        self.writer.close()


class JSONLLogger(Logger):
    def __init__(self, cfg: "TrainingConfig", log_dir: "str | PathLike"):
        super().__init__(cfg, log_dir)
        self.log_file = (self.log_dir / "logs.jsonl").open("a")

    def write(self, data: LogData):
        json_line = json.dumps(data)
        self.log_file.write(json_line + "\n")

    def close(self):
        self.log_file.close()


def from_cfg(cfg: "TrainingConfig", log_dir: "str | PathLike") -> Logger:
    if cfg.logger == "jsonl":
        return JSONLLogger(cfg, log_dir)
    elif cfg.logger == "tensorboard":
        return TensorboardLogger(cfg, log_dir)
    else:
        return WAndBLogger(cfg, log_dir)



import datetime
from typing import TYPE_CHECKING

import numpy as np
import torch
import torch.nn.functional as F
from datasets import (
    Dataset,
    IterableDataset,
)
from sklearn.metrics import confusion_matrix
from torch.utils.data import DataLoader

from saefarer.analysis.types import (
    ConfusionMatrix,
    ConfusionMatrixCell,
    DatasetInfo,
    ModelInfo,
)

if TYPE_CHECKING:
    from transformers import PreTrainedModel

    from saefarer.analysis.config import AnalysisConfig


@torch.inference_mode()
def get_dataset_with_predictions(
    model: "PreTrainedModel",
    dataset: Dataset | IterableDataset | DataLoader,
    cfg: "AnalysisConfig",
) -> dict[str, torch.Tensor]:
    print("Loading instances")

    if isinstance(dataset, Dataset):
        ds = dataset[0 : cfg.total_analysis_sequences]
    else:
        if isinstance(dataset, IterableDataset):
            dataloader = DataLoader(
                dataset,  # type: ignore
                batch_size=cfg.total_analysis_sequences,
            )
        else:
            dataloader = DataLoader(
                dataset=dataset.dataset,
                shuffle=False,
                batch_size=cfg.total_analysis_sequences,
                collate_fn=dataset.collate_fn,
                num_workers=dataset.num_workers,
            )

        ds = next(iter(dataloader))

    print("Beginning to get model's predictions on instances", datetime.datetime.now())

    predicted_probabilities = _get_model_predictions(model, ds, cfg)
    ds["pred_probs"] = predicted_probabilities
    ds["pred_label"] = predicted_probabilities.argmax(dim=1)

    print("Finished getting model's predictions on instances", datetime.datetime.now())

    return ds


def get_dataset_info(cfg: "AnalysisConfig"):
    label_indices = list(range(len(cfg.labels)))

    return DatasetInfo(
        n_sequences=cfg.total_analysis_sequences,
        n_tokens=cfg.total_analysis_sequences * cfg.model_sequence_length,
        labels=cfg.labels,
        label_indices=label_indices,
    )


@torch.inference_mode()
def _get_model_predictions(
    model: "PreTrainedModel",
    ds: dict[str, torch.Tensor],
    cfg: "AnalysisConfig",
) -> torch.Tensor:
    tokens = ds[cfg.token_ids_column]
    attn_masks = ds[cfg.attn_mask_column]

    predicted_probabilities = torch.zeros(
        (tokens.shape[0], len(cfg.labels)),
        device=torch.device("cpu"),
        dtype=model.dtype,
    )

    offset = 0

    token_batches = tokens.split(cfg.model_batch_size_sequences)
    attn_mask_batches = attn_masks.split(cfg.model_batch_size_sequences)

    for token_batch, attn_mask_batch in zip(token_batches, attn_mask_batches):
        output = model(
            token_batch.to(cfg.device),
            attention_mask=attn_mask_batch.to(cfg.device),
        )
        probs = F.softmax(output.logits, dim=1)

        start = offset
        offset += probs.shape[0]
        end = offset

        predicted_probabilities[start:end, :] = probs.to("cpu")

    return predicted_probabilities


@torch.inference_mode()
def get_model_info(
    ds: dict[str, torch.Tensor],
    dataset_info: DatasetInfo,
    cfg: "AnalysisConfig",
) -> ModelInfo:
    cm = get_confusion_matrix(
        ds[cfg.label_column], ds["pred_label"], dataset_info["label_indices"]
    )

    mean_probabilities = ds["pred_probs"].mean(dim=0).tolist()

    nll = F.nll_loss(torch.log(ds["pred_probs"]), ds[cfg.label_column])

    return ModelInfo(
        cm=cm, mean_pred_label_probs=mean_probabilities, log_loss=nll.item()
    )


@torch.inference_mode()
def get_confusion_matrix(
    y_true: torch.Tensor,
    y_pred: torch.Tensor,
    label_indices: list[int],
) -> ConfusionMatrix:
    y_true_np = y_true.numpy(force=True)
    y_pred_np = y_pred.numpy(force=True)

    n_sequences = y_true_np.shape[0]

    error_count = (y_true_np != y_pred_np).sum().item()
    error_pct = error_count / n_sequences

    label_counts = np.bincount(y_true_np, minlength=len(label_indices))
    label_pcts = label_counts / n_sequences

    pred_label_counts = np.bincount(y_pred_np, minlength=len(label_indices))
    pred_label_pcts = pred_label_counts / n_sequences

    matrix = confusion_matrix(y_true=y_true_np, y_pred=y_pred_np, labels=label_indices)

    cells: list[ConfusionMatrixCell] = []

    for true_index in label_indices:
        for pred_index in label_indices:
            count = matrix[true_index, pred_index].item()
            pct = count / n_sequences
            cell = ConfusionMatrixCell(
                label=true_index,
                pred_label=pred_index,
                count=count,
                pct=pct,
            )
            cells.append(cell)

    correct_counts = np.diagonal(matrix)

    false_pos_counts = pred_label_counts - correct_counts
    false_pos_pcts = false_pos_counts / n_sequences

    false_neg_counts = label_counts - correct_counts
    false_neg_pcts = false_neg_counts / n_sequences

    cm = ConfusionMatrix(
        n_sequences=n_sequences,
        error_count=error_count,
        error_pct=error_pct,
        cells=cells,
        label_counts=label_counts.tolist(),
        label_pcts=label_pcts.tolist(),
        pred_label_counts=pred_label_counts.tolist(),
        pred_label_pcts=pred_label_pcts.tolist(),
        false_pos_counts=false_pos_counts.tolist(),
        false_pos_pcts=false_pos_pcts.tolist(),
        false_neg_counts=false_neg_counts.tolist(),
        false_neg_pcts=false_neg_pcts.tolist(),
    )

    return cm



import math
import os
import sqlite3
from pathlib import Path
from typing import TYPE_CHECKING

import anywidget
import traitlets

import saefarer.analysis.database as db
from saefarer.analysis.config import AnalysisConfig
from saefarer.analysis.inference import inference

if TYPE_CHECKING:
    from transformers import PreTrainedModel

    from saefarer import sae
    from saefarer.analysis.types import (
        FeatureTokenSequence,
        InferenceInput,
        RankingOption,
    )
    from saefarer.protocols import TokenizerProtocol
    from saefarer.widget.config import WidgetConfig

_DEV = False


class Widget(anywidget.AnyWidget):
    if _DEV:
        _esm = "http://localhost:5173/js/widget.ts?anywidget"
        _css = ""
    else:
        bundled_assets_dir = Path(__file__).parent.parent / "static"
        _esm = bundled_assets_dir / "widget.js"
        _css = bundled_assets_dir / "saefarer.css"

    height = traitlets.Int().tag(sync=True)
    base_font_size = traitlets.Int().tag(sync=True)
    n_table_rows = traitlets.Int().tag(sync=True)

    dataset_info = traitlets.Dict().tag(sync=True)
    model_info = traitlets.Dict().tag(sync=True)

    sae_ids = traitlets.List().tag(sync=True)
    sae_id = traitlets.Unicode().tag(sync=True)
    sae_data = traitlets.Dict().tag(sync=True)

    table_ranking_option = traitlets.Dict().tag(sync=True)  # type: ignore
    table_min_act_rate = traitlets.Float().tag(sync=True)
    table_page_index = traitlets.Int().tag(sync=True)
    max_table_page_index = traitlets.Int().tag(sync=True)
    num_filtered_features = traitlets.Int().tag(sync=True)
    table_features = traitlets.List().tag(sync=True)

    detail_feature = traitlets.Dict().tag(sync=True)
    detail_feature_id = traitlets.Int().tag(sync=True)

    can_inference = traitlets.Bool().tag(sync=True)
    inference_input = traitlets.Dict().tag(sync=True)  # type: ignore
    inference_output = traitlets.Dict().tag(sync=True)  # type: ignore

    def __init__(
        self,
        path: str | os.PathLike,
        cfg: "WidgetConfig",
        model: "PreTrainedModel | None" = None,
        tokenizer: "TokenizerProtocol | None" = None,
        sae: "sae.SAE | None" = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        path = Path(path)

        if not path.exists():
            raise OSError(f"Cannot read {path}")

        self.cfg = cfg

        self.con = sqlite3.connect(path.as_posix())
        self.cur = self.con.cursor()

        self.height = cfg.height
        self.base_font_size = cfg.base_font_size
        self.n_table_rows = cfg.n_table_rows

        analysis_cfg_dict = db.read_misc("analysis_cfg", self.cur)
        self.analysis_cfg = AnalysisConfig.from_dict(analysis_cfg_dict)
        self.dataset_info = db.read_misc("dataset_info", self.cur)
        self.model_info = db.read_misc("model_info", self.cur)

        self.sae_ids = db.read_sae_ids(self.cur)
        self.sae_id = self.sae_ids[0]
        self.sae_data = db.read_sae_data(self.sae_ids[0], self.cur)

        self.table_ranking_option: "RankingOption" = {
            "kind": "label",
            "true_label": "different",
            "pred_label": "any",
            "descending": True,
        }
        self.table_min_act_rate = (
            cfg.default_min_act_rate
            if cfg.default_min_act_rate is not None
            else self.get_default_min_act_rate()
        )
        self.table_page_index = 0
        self.num_filtered_features = self.sae_data["n_alive_features"]
        self.max_table_page_index = (
            math.ceil(self.num_filtered_features / self.n_table_rows) - 1
        )
        self.table_features = db.rank_features(
            self.sae_id,
            self.cur,
            self.table_ranking_option,
            self.table_min_act_rate,
            self.table_page_index,
            self.n_table_rows,
            len(self.dataset_info["labels"]),
        )

        self.detail_feature = self.table_features[0]
        self.detail_feature_id = self.detail_feature["feature_id"]

        self.model = model
        self.tokenizer = tokenizer
        self.sae = sae

        if (
            self.model is not None
            and self.tokenizer is not None
            and self.sae is not None
        ):
            self.can_inference = True
            self.model.to(cfg.device)  # type: ignore
        else:
            self.can_inference = False

        if not self.can_inference and (
            model is not None or tokenizer is not None or sae is not None
        ):
            raise ValueError(
                "model, tokenizer, and sae must all be set in order to inference"
            )

        self.inference_input: "InferenceInput" = {
            "feature_index": -1,
            "sequence": "",
        }
        self.inference_output: "FeatureTokenSequence" = {
            "feature_index": -1,
            "sequence_index": -1,
            "display_tokens": [],
            "max_token_index": -1,
            "label": -1,
            "pred_label": -1,
            "pred_probs": [],
            "extras": {},
        }

    @traitlets.observe("detail_feature_id")
    def _on_detail_feature_id_change(self, change):
        new_feature_id = change["new"]

        # this happens when we reset detail_feature_id to the id of
        # detail_feature when an invalid value is passed
        if new_feature_id == self.detail_feature["feature_id"]:
            return

        for feature in self.table_features:
            if new_feature_id == feature["feature_id"]:
                self.detail_feature = feature
                return

        feature = db.read_feature_data(new_feature_id, self.sae_id, self.cur)

        if feature is not None:
            self.detail_feature = feature
            return

        self.detail_feature_id = self.detail_feature["feature_id"]

    @traitlets.observe("table_page_index")
    def _on_table_page_index_change(self, _):
        self._update_table_features()

    @traitlets.observe("table_ranking_option")
    def table_ranking_option_change(self, _):
        """When the ranking option is changed, go back to the first page.
        Updating table_features will happen in the change handler for
        table_page_index. If we are already on the first page,
        then update table_features here."""

        if self.table_page_index == 0:
            self._update_table_features()
        else:
            self.table_page_index = 0

    @traitlets.observe("table_min_act_rate")
    def _on_table_min_act_rate(self, _):
        """When the table is filtered, go back to the first page.
        Updating table_features will happen in the change handler for
        table_page_index. If we are already on the first page,
        then update table_features here."""

        self.num_filtered_features = db.count_features(
            self.sae_id,
            self.cur,
            self.table_min_act_rate,
        )
        self.max_table_page_index = (
            math.ceil(self.num_filtered_features / self.n_table_rows) - 1
        )

        if self.table_page_index == 0:
            self._update_table_features()
        else:
            self.table_page_index = 0

    def _update_table_features(self):
        self.table_features = db.rank_features(
            self.sae_id,
            self.cur,
            self.table_ranking_option,
            self.table_min_act_rate,
            self.table_page_index,
            self.n_table_rows,
            len(self.dataset_info["labels"]),
        )

    @traitlets.observe("inference_input")
    def _on_inference_input_change(self, _):
        if self.inference_input["feature_index"] == -1:
            return

        if not self.can_inference:
            return

        if self.model is None or self.tokenizer is None or self.sae is None:
            return

        self.inference_output = inference(
            self.inference_input,
            self.model,
            self.tokenizer,
            self.sae,
            self.analysis_cfg,
        )

    def get_default_min_act_rate(self):
        raw = (
            self.cfg.approx_default_min_act_instances / self.dataset_info["n_sequences"]
        )

        if raw < 0:
            return 0

        if raw > 1:
            return 1

        places = abs(math.floor(math.log10(raw)))
        return round(raw, places)



"""Code for training a sparse autoencoder."""

import time
from pathlib import Path
from typing import TYPE_CHECKING

import torch
import tqdm
from torch.utils.data import DataLoader

from saefarer.sae import SAE
from saefarer.training import logger
from saefarer.training.activations_store import ActivationsStore

if TYPE_CHECKING:
    from os import PathLike

    from datasets import (
        Dataset,
        IterableDataset,
    )
    from transformers import PreTrainedModel

    from saefarer.sae import ForwardOutput
    from saefarer.training.config import TrainingConfig


def train(
    cfg: "TrainingConfig",
    model: "PreTrainedModel",
    dataset: "Dataset | IterableDataset | DataLoader",
    save_path: "str | PathLike",
    log_dir: "str | PathLike",
    checkpoint_dir: "str | PathLike | None" = None,
) -> SAE:
    """Train the SAE"""

    log = logger.from_cfg(cfg, log_dir)

    sae = SAE(cfg)

    store = ActivationsStore(model, dataset, cfg)

    optimizer = torch.optim.Adam(
        sae.parameters(),
        lr=cfg.lr,
        betas=(cfg.beta1, cfg.beta2),
        eps=cfg.eps,
    )

    print("Beginning training")

    start_time = time.time()

    for i in tqdm.trange(
        1, cfg.total_training_batches + 1, disable=not cfg.show_progress
    ):
        # get next batch of model activations
        x = store.next_batch()

        # forward pass through SAE
        output: "ForwardOutput" = sae(x)

        # backward pass

        output.loss.backward()

        sae.set_decoder_norm_to_unit_norm()
        sae.remove_gradient_parallel_to_decoder_directions()

        optimizer.step()
        optimizer.zero_grad()

        # logging

        if cfg.log_batch_freq and (
            i % cfg.log_batch_freq == 0 or i == cfg.total_training_batches
        ):
            log_data = logger.LogData(
                elapsed_seconds=time.time() - start_time,
                n_training_batches=i,
                n_training_tokens=i * sae.cfg.sae_batch_size_tokens,
                loss=output.loss.item(),
                mse_loss=output.mse_loss.item(),
                aux_loss=output.aux_loss.item(),
                n_dead_features=output.num_dead,
                mean_n_batches_since_fired=sae.stats_last_nonzero.mean(
                    dtype=torch.float32
                ).item(),
                max_n_batches_since_fired=int(sae.stats_last_nonzero.max().item()),
            )

            log.write(log_data)

        if (
            cfg.checkpoint_batch_freq
            and checkpoint_dir
            and i % cfg.checkpoint_batch_freq == 0
        ):
            print(f"Saving checkpoint after batch {i}")
            sae.save(Path(checkpoint_dir) / f"batch_{i}.pt")

    print("Saving final model")

    sae.save(save_path)

    log.close()

    return sae



"""
This is based on `activations_store.py` from SAELens.
"""

from typing import TYPE_CHECKING, Any, Iterator

import torch
from datasets import Dataset, IterableDataset
from einops import rearrange
from torch.utils.data import DataLoader, TensorDataset

from saefarer.constants import DTYPES

if TYPE_CHECKING:
    from transformers import PreTrainedModel

    from saefarer.training.config import TrainingConfig


class ActivationsStore:
    """
    This class is used to provide model activations from a given
    layer to train the SAE on.
    """

    def __init__(
        self,
        model: "PreTrainedModel",
        dataset: Dataset | IterableDataset | DataLoader,
        cfg: "TrainingConfig",
    ):
        self.dtype = DTYPES[cfg.dtype]
        self.device = torch.device(cfg.device)

        self.model = model.to(self.device)  # type: ignore

        self.cfg = cfg

        self._activations_dataloader: Iterator[Any] | None = None
        self._activations_storage_buffer: torch.Tensor | None = None

        if isinstance(dataset, Dataset) or isinstance(dataset, IterableDataset):
            self.dataset_dataloader = DataLoader(
                dataset,  # type: ignore
                batch_size=self.cfg.model_batch_size_sequences,
                drop_last=True,
            )
        else:
            self.dataset_dataloader = dataset

        batch_shape = next(iter(self.dataset_dataloader))[
            self.cfg.token_ids_column
        ].shape

        assert batch_shape[0] == self.cfg.model_batch_size_sequences, (
            f"DataLoader batch size is {batch_shape[0]} but cfg.model_batch_size_sequences = {self.cfg.model_batch_size_sequences}"
        )

        assert batch_shape[1] == self.cfg.model_sequence_length, (
            f"Dataset sequence length is {batch_shape[1]} but cfg.model_sequence_length = {self.cfg.model_sequence_length}"
        )

        self.dataset_batch_iter = iter(self.dataset_dataloader)
        self.num_samples_processed = 0

    @property
    def activations_storage_buffer(self) -> torch.Tensor:
        """
        The storage buffer contains half of the activations in the store.
        It is used to refill the dataloader when it runs out.
        """
        if self._activations_storage_buffer is None:
            self._activations_storage_buffer = self.get_buffer(
                self.cfg.n_batches_in_store // 2
            )

        return self._activations_storage_buffer

    @property
    def activations_dataloader(self) -> Iterator[Any]:
        """
        The dataloader contains half of the activations in the store
        and is iterated over to get batches of activations.
        When it runs out, more activations are retrieved and get shuffled
        with the storage buffer.
        """
        if self._activations_dataloader is None:
            self._activations_dataloader = self.get_activations_data_loader()

        return self._activations_dataloader

    @torch.no_grad()
    def get_buffer(self, n_batches, raise_at_epoch_end: bool = False) -> torch.Tensor:
        """Get buffer of activations."""
        n_tokens_in_model_batch = (
            self.cfg.model_batch_size_sequences * self.cfg.model_sequence_length
        )
        n_tokens_in_buffer = n_batches * n_tokens_in_model_batch

        new_buffer = torch.zeros(
            (n_tokens_in_buffer, self.cfg.d_in),
            dtype=self.dtype,
            requires_grad=False,
            device=self.device,
        )

        n_tokens_read = 0

        while n_tokens_read < n_tokens_in_buffer:
            tokens, attn_mask = self.get_batch_tokens(raise_at_epoch_end)
            activations = self.get_activations(tokens, attn_mask)

            n_tokens = activations.shape[0]
            # TODO: don't drop unused tokens
            n_tokens_to_use = min(n_tokens, n_tokens_in_buffer - n_tokens_read)

            start = n_tokens_read
            end = start + n_tokens_to_use
            new_buffer[start:end] = activations[0:n_tokens_to_use]

            n_tokens_read += n_tokens_to_use

        new_buffer = new_buffer[torch.randperm(new_buffer.shape[0])]

        return new_buffer

    def get_activations_data_loader(self) -> Iterator[Any]:
        """Create new dataloader."""
        try:
            new_samples = self.get_buffer(
                self.cfg.n_batches_in_store // 2, raise_at_epoch_end=True
            )
        except StopIteration as e:
            # Dump current buffer so that samples aren't leaked between epochs
            self._activations_storage_buffer = None

            print(str(e))

            try:
                new_samples = self.get_buffer(
                    self.cfg.n_batches_in_store // 2, raise_at_epoch_end=True
                )
            except StopIteration:
                raise ValueError("Unable to fill buffer after starting new epoch.")

        mixing_buffer = torch.cat([new_samples, self.activations_storage_buffer], dim=0)
        mixing_buffer = mixing_buffer[torch.randperm(mixing_buffer.shape[0])]

        self._activations_storage_buffer = mixing_buffer[: mixing_buffer.shape[0] // 2]

        dataset = TensorDataset(mixing_buffer[mixing_buffer.shape[0] // 2 :])

        # TODO: check that batch size is smaller than dataset size
        dataloader = DataLoader(
            dataset,
            batch_size=self.cfg.sae_batch_size_tokens,
            shuffle=True,
        )

        dataloader_iterator = iter(dataloader)

        return dataloader_iterator

    def get_batch_tokens(
        self, raise_at_epoch_end: bool = False
    ) -> tuple[torch.Tensor, torch.Tensor | None]:
        """Get batch of tokens from the dataset."""

        def get_tokens_and_attn_mask():
            batch = next(self.dataset_batch_iter)
            tokens = batch[self.cfg.token_ids_column].to(self.device)
            mask = (
                batch[self.cfg.attn_mask_column].to(self.device)
                if self.cfg.attn_mask_column
                else None
            )
            self.num_samples_processed += self.cfg.model_batch_size_sequences
            return tokens, mask

        try:
            return get_tokens_and_attn_mask()
        except StopIteration:
            self.dataset_batch_iter = iter(self.dataset_dataloader)

            if raise_at_epoch_end:
                raise StopIteration(
                    f"Ran out of tokens in dataset after {self.num_samples_processed} samples."
                )
            else:
                return get_tokens_and_attn_mask()

    @torch.no_grad()
    def get_activations(
        self, batch_tokens: torch.Tensor, attn_mask: torch.Tensor | None
    ) -> torch.Tensor:
        """Get activations for tokens."""
        batch_output = self.model(
            batch_tokens, attention_mask=attn_mask, output_hidden_states=True
        )
        batch_activations = batch_output.hidden_states[self.cfg.hidden_state_index]
        flat_activations = rearrange(
            batch_activations, "batches seq_len d_in -> (batches seq_len) d_in"
        )

        if attn_mask is not None:
            flat_attn_mask = rearrange(
                attn_mask, "batches seq_len -> (batches seq_len)"
            ).bool()

            flat_activations = flat_activations[flat_attn_mask]

        return flat_activations

    def next_batch(self):
        """Get batch of activations."""
        try:
            return next(self.activations_dataloader)[0]
        except StopIteration:
            # if the dataloader is exhausted, create a new one
            self._activations_dataloader = self.get_activations_data_loader()
            return next(self.activations_dataloader)[0]



"""Configuration for SAE training."""

from dataclasses import dataclass, field
from typing import Literal


@dataclass
class TrainingConfig:
    """Configuration class for training SAEs."""

    device: Literal["cpu", "mps", "cuda", "xpu", "xla"] = "cuda"
    dtype: Literal["float16", "bfloat16", "float32", "float64"] = "float32"
    # dataset
    token_ids_column: str = "input_ids"
    attn_mask_column: str = "attention_mask"
    # dimensions
    d_in: int = 64
    expansion_factor: int = 4
    d_sae: int = field(init=False)
    # loss
    k: int = 4
    aux_k: int = 32
    aux_k_coef: float = 1 / 32
    dead_tokens_threshold: int = 10_000_000
    dead_steps_threshold: int = field(init=False)
    # inferencing
    hidden_state_index: int = -2
    # activation normalization
    normalize: bool = False
    # batch sizes
    model_sequence_length: int = 256
    model_batch_size_sequences: int = 32
    n_batches_in_store: int = 20
    sae_batch_size_tokens: int = 4096
    # adam
    lr: float = 3e-4
    beta1: float = 0.9
    beta2: float = 0.999
    eps: float = 6.25e-10
    # training
    total_training_tokens: int = 100_000_000
    total_training_batches: int = field(init=False)
    # logging
    show_progress: bool = True
    logger: Literal["jsonl", "wandb", "tensorboard"] = "jsonl"
    log_batch_freq: int = 1000
    wandb_project: str | None = None
    wandb_group: str | None = None
    wandb_name: str | None = None
    wandb_notes: str | None = None
    # checkpointing
    checkpoint_batch_freq: int = 10_000

    def __post_init__(self):
        self.dead_steps_threshold = (
            self.dead_tokens_threshold // self.sae_batch_size_tokens
        )

        self.total_training_batches = (
            self.total_training_tokens // self.sae_batch_size_tokens
        )

        self.d_sae = self.d_in * self.expansion_factor



"""Configuration for Widget."""

from dataclasses import dataclass
from typing import Literal


@dataclass
class WidgetConfig:
    """Configuration class for the widget."""

    height: int = 600
    base_font_size: int = 16
    n_table_rows: int = 10
    approx_default_min_act_instances: int | None = 100
    default_min_act_rate: float | None = None
    # only for inference
    device: 'Literal["cpu", "mps", "cuda", "xpu", "xla"]' = "cuda"



from dataclasses import dataclass
from typing import Literal, TypedDict

import torch

# Aligned with types.ts


class HistogramData(TypedDict):
    counts: list[int]
    thresholds: list[float]


class MarginalEffectsData(TypedDict):
    probs: list[list[float]]
    thresholds: list[float]
    non_act_probs: list[float]


class DisplayToken(TypedDict):
    display: str
    token_ids: list[int]
    acts: list[float]
    max_act: float
    extras: dict[str, list[str]]
    is_padding: bool


class FeatureTokenSequence(TypedDict):
    feature_index: int
    sequence_index: int
    display_tokens: list[DisplayToken]
    max_token_index: int
    label: int
    pred_label: int
    pred_probs: list[float]
    extras: dict[str, str]


class SequenceInterval(TypedDict):
    min_max_act: float
    max_max_act: float
    sequences: list[FeatureTokenSequence]


class ConfusionMatrixCell(TypedDict):
    label: int
    pred_label: int
    count: int
    pct: float


class ConfusionMatrix(TypedDict):
    n_sequences: int
    error_count: int
    error_pct: float
    cells: list[ConfusionMatrixCell]
    label_counts: list[int]
    label_pcts: list[float]
    pred_label_counts: list[int]
    pred_label_pcts: list[float]
    false_pos_counts: list[int]
    false_pos_pcts: list[float]
    false_neg_counts: list[int]
    false_neg_pcts: list[float]


class FeatureData(TypedDict):
    sae_id: str
    feature_id: int
    max_act: float
    token_act_rate: float
    token_acts_histogram: HistogramData
    sequence_act_rate: float
    sequence_acts_histogram: HistogramData
    marginal_effects: MarginalEffectsData
    cm: ConfusionMatrix
    sequence_intervals: list[SequenceInterval]
    mean_pred_label_probs: list[float]


class FeatureProjection(TypedDict):
    feature_ids: list[int]
    xs: list[float]
    ys: list[float]


class SAEData(TypedDict):
    sae_id: str
    n_total_features: int
    n_alive_features: int
    n_dead_features: int
    n_non_activating_features: int
    alive_feature_ids: list[int]
    token_act_rate_histogram: HistogramData
    sequence_act_rate_histogram: HistogramData
    feature_projection: FeatureProjection


class FeatureIdRankingOption(TypedDict):
    kind: Literal["feature_id"]
    descending: bool


class SequenceActRateRankingOption(TypedDict):
    kind: Literal["sequence_act_rate"]
    descending: bool


class LabelRankingOption(TypedDict):
    kind: Literal["label"]
    true_label: str
    pred_label: str
    descending: bool


RankingOption = (
    FeatureIdRankingOption | SequenceActRateRankingOption | LabelRankingOption
)


class DatasetInfo(TypedDict):
    labels: list[str]
    label_indices: list[int]
    n_sequences: int
    n_tokens: int


class ModelInfo(TypedDict):
    cm: ConfusionMatrix
    mean_pred_label_probs: list[float]
    log_loss: float


class InferenceInput(TypedDict):
    feature_index: int
    sequence: str


# Python only


@dataclass
class SequenceIntervalIndices:
    min_max_act: float
    max_max_act: float
    indices: torch.Tensor



"""
Sparse autoencoder model.
This is based on code from OpenAI:
https://github.com/openai/sparse_autoencoder
"""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Literal

import einops
import torch
import torch.nn as nn
import torch.nn.functional as F

if TYPE_CHECKING:
    from os import PathLike

    from typing_extensions import Self

    from saefarer.training.config import TrainingConfig


@dataclass
class ForwardOutput:
    """SAE forward output"""

    reconstructions: torch.Tensor
    loss: torch.Tensor
    mse_loss: torch.Tensor
    aux_loss: torch.Tensor
    num_dead: int


def normalized_mse(output, target):
    """Normalized MSE loss"""
    target_mu = target.mean(dim=0)
    target_mu_reshaped = target_mu.unsqueeze(0).broadcast_to(target.shape)
    mse = F.mse_loss(output, target, reduction="mean")
    mse_naive = F.mse_loss(target_mu_reshaped, target, reduction="mean")
    return mse / mse_naive


def LN(
    x: torch.Tensor, eps: float = 1e-5
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    mu = x.mean(dim=-1, keepdim=True)
    x = x - mu
    std = x.std(dim=-1, keepdim=True)
    x = x / (std + eps)
    return x, mu, std


class SAE(nn.Module):
    """Sparse autoencoder

    Implements:
        latents = relu(topk(encoder(x - b_dec) + b_enc))
        recons = decoder(latents) + b_dec
    """

    def __init__(self, cfg: "TrainingConfig") -> None:
        super().__init__()

        self.dtype = getattr(torch, cfg.dtype)
        self.device = torch.device(cfg.device)

        self.cfg = cfg

        self.b_enc = nn.Parameter(torch.zeros(cfg.d_sae, device=self.device))
        self.b_dec = nn.Parameter(torch.zeros(cfg.d_in, device=self.device))

        self.W_enc = nn.Parameter(
            torch.nn.init.kaiming_uniform_(
                torch.empty(
                    cfg.d_in,
                    cfg.d_sae,
                    dtype=self.dtype,
                    device=self.device,
                )
            )
        )

        # each row of W_dec corresponds to one feature of the SAE
        # each col corresponds to one neuron in the model
        self.W_dec = nn.Parameter(self.W_enc.t().clone())

        self.set_decoder_norm_to_unit_norm()

        self.topk = TopK(cfg.k)
        self.aux_topk = TopK(cfg.aux_k)

        self.stats_last_nonzero: torch.Tensor
        self.register_buffer(
            "stats_last_nonzero",
            torch.zeros(cfg.d_sae, dtype=torch.long, device=self.device),
        )

    def get_dead_neuron_mask(self) -> torch.Tensor:
        return self.stats_last_nonzero > self.cfg.dead_steps_threshold

    def auxk_masker(self, x: torch.Tensor) -> torch.Tensor:
        """mask dead neurons"""
        dead_mask = self.get_dead_neuron_mask()
        x.data *= dead_mask
        return x

    def encode_pre_act(self, x: torch.Tensor) -> torch.Tensor:
        """
        :param x: input data (shape: [batch, n_inputs])
        :return: autoencoder latents before activation (shape: [batch, n_latents])
        """
        x = x - self.b_dec
        latents_pre_act = x @ self.W_enc + self.b_enc
        return latents_pre_act

    def preprocess(
        self, x: torch.Tensor
    ) -> tuple[torch.Tensor, dict[str, torch.Tensor]]:
        """Mean center, standard"""
        if not self.cfg.normalize:
            return x, dict()
        x, mu, std = LN(x)
        return x, dict(mu=mu, std=std)

    def unprocess(
        self, x: torch.Tensor, info: dict[str, torch.Tensor] | None = None
    ) -> torch.Tensor:
        if self.cfg.normalize and info:
            return x * info["std"] + info["mu"]
        else:
            return x

    def encode(self, x: torch.Tensor) -> tuple[torch.Tensor, dict[str, torch.Tensor]]:
        """
        :param x: input data (shape: [batch, n_inputs])
        :return: autoencoder latents (shape: [batch, n_latents])
        """
        x, info = self.preprocess(x)
        return self.topk(self.encode_pre_act(x)), info

    def decode(
        self, latents: torch.Tensor, info: dict[str, Any] | None = None
    ) -> torch.Tensor:
        """
        :param latents: autoencoder latents (shape: [batch, n_latents])
        :return: reconstructed data (shape: [batch, n_inputs])
        """
        recontructed = latents @ self.W_dec + self.b_dec
        return self.unprocess(recontructed, info)

    def forward(self, x: torch.Tensor) -> ForwardOutput:
        """
        :param x: input data (shape: [batch, n_inputs])
        :return:  autoencoder latents pre activation (shape: [batch, n_latents])
                  autoencoder latents (shape: [batch, n_latents])
                  reconstructed data (shape: [batch, n_inputs])
        """
        x_preprocessed, info = self.preprocess(x)
        latents_pre_act = self.encode_pre_act(x_preprocessed)
        latents = self.topk(latents_pre_act)
        # not passing info to decode so that the reconstructed
        # activations are still centered
        # TODO: would it be better uncenter the reconstructed activations
        # and then compare against x rather than x_preprocessed?
        recons = self.decode(latents)

        mse_loss = normalized_mse(recons, x_preprocessed)

        # set all indices of self.stats_last_nonzero where (latents != 0) to 0
        self.stats_last_nonzero *= (latents == 0).all(dim=0).long()
        self.stats_last_nonzero += 1

        num_dead = int(self.get_dead_neuron_mask().sum().item())

        if num_dead > 0:
            aux_latents = self.aux_topk(self.auxk_masker(latents_pre_act))
            aux_recons = self.decode(aux_latents)
            aux_loss = self.cfg.aux_k_coef * normalized_mse(
                aux_recons, x_preprocessed - recons.detach() + self.b_dec.detach()
            ).nan_to_num(0)
        else:
            aux_loss = mse_loss.new_tensor(0.0)

        loss = mse_loss + aux_loss

        return ForwardOutput(
            reconstructions=self.unprocess(recons),
            loss=loss,
            mse_loss=mse_loss,
            aux_loss=aux_loss,
            num_dead=num_dead,
        )

    @torch.no_grad()
    def set_decoder_norm_to_unit_norm(self):
        """
        Set decoder weights to have unit norm.
        """
        self.W_dec.data /= self.W_dec.data.norm(dim=1, keepdim=True)

    @torch.no_grad()
    def remove_gradient_parallel_to_decoder_directions(self):
        """
        Update grads so that they remove the parallel component
            (d_sae, d_in) shape
        """
        assert self.W_dec.grad is not None  # keep pyright happy

        # The below code is equivalent to
        # parallel_component = (self.W_dec.grad * self.W_dec.data).sum(
        #     dim=1, keepdim=True
        # )
        # self.W_dec.grad -= parallel_component * self.W_dec.data

        parallel_component = einops.einsum(
            self.W_dec.grad,
            self.W_dec.data,
            "d_sae d_in, d_sae d_in -> d_sae",
        )

        self.W_dec.grad -= einops.einsum(
            parallel_component,
            self.W_dec.data,
            "d_sae, d_sae d_in -> d_sae d_in",
        )

    def save(self, path: "str | PathLike"):
        """Save model to path."""
        torch.save([self.cfg, self.state_dict()], path)

    @classmethod
    def load(
        cls,
        path: "str | PathLike",
        device: Literal["cpu", "mps", "cuda", "xpu", "xla"] | torch.device,
    ) -> "Self":
        """Load model from path."""
        config, state = torch.load(path, map_location=device, weights_only=False)

        config.device = device

        model = cls(config)
        model.load_state_dict(state)

        return model


class TopK(nn.Module):
    """TopK activation."""

    def __init__(self, k: int, postact_fn: Callable = nn.ReLU()) -> None:
        super().__init__()
        self.k = k
        self.postact_fn = postact_fn

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass."""
        topk = torch.topk(x, k=self.k, dim=-1)
        values = self.postact_fn(topk.values)
        # make all other values 0
        result = torch.zeros_like(x)
        result.scatter_(-1, topk.indices, values)
        return result



from typing import Protocol, TypedDict


class EncodingOutput(TypedDict):
    token_ids: list[int]
    attention_mask: list[int]


class TokenizerProtocol(Protocol):
    pad_token_id: int

    def encode(
        self,
        text: str,
        max_length: int,
    ) -> EncodingOutput: ...

    def decode(self, token_ids: list[int]) -> str: ...



"""Utility functions."""

from typing import TYPE_CHECKING, Literal

import numpy as np
import torch

if TYPE_CHECKING:
    import numpy.typing as npt


def get_default_device() -> 'Literal["cpu", "mps", "cuda"]':
    if torch.cuda.is_available():
        return "cuda"
    elif torch.backends.mps.is_available() and torch.backends.mps.is_built():
        return "mps"
    else:
        return "cpu"


def top_k_indices_values(
    x: torch.Tensor, k: int, largest: bool = True
) -> tuple[torch.Tensor, torch.Tensor]:
    """Given a 2D matrix x, return the row and column indices of
    the k largest or smallest values."""
    # This code is adapted from sae-vis.
    top = x.flatten().topk(k=k, largest=largest)
    indices = top.indices
    rows = indices // x.size(1)
    cols = indices % x.size(1)
    return torch.stack((rows, cols), dim=1), top.values


def freedman_diaconis_np(
    x: "npt.NDArray", x_range: tuple[float, float] | None = None
) -> int:
    """Freedman Diaconis Estimator for determining
    the number of bins in a histogram."""
    iqr = np.quantile(x, 0.75) - np.quantile(x, 0.25)
    bin_width = 2 * iqr / np.cbrt(x.size)

    if bin_width == 0:
        return 1

    diff = x.max() - x.min() if x_range is None else x_range[1] - x_range[0]

    n_bins = diff / bin_width
    return int(np.ceil(n_bins))



import json
import sqlite3
from typing import TYPE_CHECKING, Any, Literal, Mapping

from saefarer.analysis.types import (
    FeatureData,
    LabelRankingOption,
    RankingOption,
    SAEData,
)

if TYPE_CHECKING:
    from pathlib import Path


def create_database(output_path: "Path") -> tuple[sqlite3.Connection, sqlite3.Cursor]:
    con = sqlite3.connect(output_path.as_posix())
    cur = con.cursor()

    cur.execute("""
        CREATE TABLE misc(
            key STRING PRIMARY KEY,
            value TEXT
        )
    """)

    cur.execute("""
        CREATE TABLE sae(
            sae_id STRING PRIMARY KEY,
            n_total_features INTEGER,
            n_alive_features INTEGER,
            n_dead_features INTEGER,
            n_non_activating_features INTEGER,
            alive_feature_ids TEXT,
            token_act_rate_histogram TEXT,
            sequence_act_rate_histogram TEXT,
            feature_projection TEXT
        )
    """)

    cur.execute("""
        CREATE TABLE feature(
            sae_id TEXT,
            feature_id INTEGER,
            max_act REAL,
            token_act_rate REAL,    
            token_acts_histogram TEXT,
            sequence_act_rate REAL,    
            sequence_acts_histogram TEXT,
            marginal_effects TEXT,
            cm TEXT,
            sequence_intervals TEXT,
            mean_pred_label_probs TEXT,
            PRIMARY KEY (sae_id, feature_id)
        )
    """)

    return con, cur


def insert_misc(key: str, value: Any, con: sqlite3.Connection, cur: sqlite3.Cursor):
    cur.execute(
        """
        INSERT INTO misc VALUES(
            :key,
            :value
        )
        """,
        {"key": key, "value": json.dumps(value)},
    )
    con.commit()


def insert_sae(data: SAEData, con: sqlite3.Connection, cur: sqlite3.Cursor):
    cur.execute(
        """
        INSERT INTO sae VALUES(
            :sae_id,
            :n_total_features,
            :n_alive_features,
            :n_dead_features,
            :n_non_activating_features,
            :alive_feature_ids,
            :token_act_rate_histogram,
            :sequence_act_rate_histogram,
            :feature_projection
        )
        """,
        convert_dict_for_db(data),
    )
    con.commit()


def insert_feature(data: FeatureData, con: sqlite3.Connection, cur: sqlite3.Cursor):
    cur.execute(
        """
        INSERT INTO feature VALUES(
            :sae_id,
            :feature_id,
            :max_act,
            :token_act_rate,
            :token_acts_histogram,
            :sequence_act_rate,
            :sequence_acts_histogram,
            :marginal_effects,
            :cm,
            :sequence_intervals,
            :mean_pred_label_probs
        )
        """,
        convert_dict_for_db(data),
    )
    con.commit()


def convert_dict_for_db(x: Mapping[str, Any]) -> dict[str, Any]:
    return {
        k: v if isinstance(v, (int, float, str)) else json.dumps(v)
        for k, v in x.items()
    }


def read_misc(key: str, cur: sqlite3.Cursor) -> Any:
    res = cur.execute(
        """
        SELECT * FROM misc WHERE key = ?
        """,
        (key,),
    )
    return json.loads(res.fetchone()[1])


def read_sae_ids(cur: sqlite3.Cursor) -> list[str]:
    res = cur.execute(
        """
        SELECT sae_id FROM sae
        """
    )
    rows = res.fetchall()
    return [row[0] for row in rows]


def read_sae_data(sae_id: str, cur: sqlite3.Cursor) -> SAEData:
    res = cur.execute(
        """
        SELECT * FROM sae WHERE sae_id = ?
        """,
        (sae_id,),
    )
    (
        sae_id,
        n_total_features,
        n_alive_features,
        n_dead_features,
        n_non_activating_features,
        alive_feature_ids,
        token_act_rate_histogram,
        sequence_act_rate_histogram,
        feature_projection,
    ) = res.fetchone()

    return SAEData(
        sae_id=sae_id,
        n_total_features=n_total_features,
        n_alive_features=n_alive_features,
        n_dead_features=n_dead_features,
        n_non_activating_features=n_non_activating_features,
        alive_feature_ids=json.loads(alive_feature_ids),
        token_act_rate_histogram=json.loads(token_act_rate_histogram),
        sequence_act_rate_histogram=json.loads(sequence_act_rate_histogram),
        feature_projection=json.loads(feature_projection),
    )


def row_to_feature_data(row: Any) -> FeatureData:
    (
        sae_id,
        feature_id,
        max_act,
        token_act_rate,
        token_acts_histogram,
        sequence_act_rate,
        sequence_acts_histogram,
        marginal_effects,
        cm,
        sequence_intervals,
        mean_pred_label_probs,
    ) = row

    return FeatureData(
        sae_id=sae_id,
        feature_id=feature_id,
        max_act=max_act,
        token_act_rate=token_act_rate,
        token_acts_histogram=json.loads(token_acts_histogram),
        sequence_act_rate=sequence_act_rate,
        sequence_acts_histogram=json.loads(sequence_acts_histogram),
        marginal_effects=json.loads(marginal_effects),
        cm=json.loads(cm),
        sequence_intervals=json.loads(sequence_intervals),
        mean_pred_label_probs=json.loads(mean_pred_label_probs),
    )


def read_feature_data(
    feature_id: int, sae_id: str, cur: sqlite3.Cursor
) -> FeatureData | None:
    res = cur.execute(
        """
        SELECT * FROM feature WHERE sae_id = ? AND feature_id = ? 
        """,
        (
            sae_id,
            feature_id,
        ),
    )

    row = res.fetchone()

    if row is None:
        return None

    return row_to_feature_data(row)


def rank_features(
    sae_id: str,
    cur: sqlite3.Cursor,
    ranking_option: RankingOption,
    min_act_rate: float,
    page_index: int,
    n_table_rows: int,
    n_labels: int,
) -> list[FeatureData]:
    if ranking_option["kind"] == "label":
        res = _rank_features_by_label(
            sae_id=sae_id,
            cur=cur,
            ranking_option=ranking_option,
            min_act_rate=min_act_rate,
            page_index=page_index,
            n_table_rows=n_table_rows,
            n_labels=n_labels,
        )
    elif (
        ranking_option["kind"] == "feature_id"
        or ranking_option["kind"] == "sequence_act_rate"
    ):
        res = _rank_features_by_col(
            sae_id=sae_id,
            cur=cur,
            col=ranking_option["kind"],
            is_descending=ranking_option["descending"],
            min_act_rate=min_act_rate,
            page_index=page_index,
            n_table_rows=n_table_rows,
        )
    else:
        raise ValueError("Unknown kind")

    rows = res.fetchall()

    return [row_to_feature_data(row) for row in rows]


def count_features(
    sae_id: str,
    cur: sqlite3.Cursor,
    min_act_rate: float,
) -> int:
    res = cur.execute(
        """
        SELECT COUNT(feature_id)
        FROM feature
        WHERE sae_id = :sae_id AND sequence_act_rate > :min_act_rate
        """,
        {
            "sae_id": sae_id,
            "min_act_rate": min_act_rate,
        },
    )

    row = res.fetchone()

    return row[0]


def _rank_features_by_col(
    sae_id: str,
    cur: sqlite3.Cursor,
    col: Literal["feature_id"] | Literal["sequence_act_rate"],
    is_descending: bool,
    min_act_rate: float,
    page_index: int,
    n_table_rows: int,
) -> sqlite3.Cursor:
    return cur.execute(
        f"""
        SELECT *
        FROM feature
        WHERE sae_id = :sae_id AND sequence_act_rate > :min_act_rate
        ORDER BY {col} {"DESC" if is_descending else "ASC"}
        LIMIT :limit
        OFFSET :offset
        """,
        {
            "sae_id": sae_id,
            "min_act_rate": min_act_rate,
            "limit": n_table_rows,
            "offset": n_table_rows * page_index,
        },
    )


def _rank_features_by_label(
    sae_id: str,
    cur: sqlite3.Cursor,
    ranking_option: LabelRankingOption,
    min_act_rate: float,
    page_index: int,
    n_table_rows: int,
    n_labels: int,
) -> sqlite3.Cursor:
    y_true = ranking_option["true_label"]
    y_pred = ranking_option["pred_label"]
    is_descending = ranking_option["descending"]

    if (y_true == "any" and y_pred == "any") or (
        y_true == "different" and y_pred == "different"
    ):
        return _rank_features_by_col(
            sae_id=sae_id,
            cur=cur,
            col="feature_id",
            is_descending=is_descending,
            min_act_rate=min_act_rate,
            page_index=page_index,
            n_table_rows=n_table_rows,
        )
    elif (y_true == "any" and y_pred == "different") or (
        y_true == "different" and y_pred == "any"
    ):
        return _rank_features_by_overall_error_pct(
            sae_id=sae_id,
            cur=cur,
            is_descending=is_descending,
            min_act_rate=min_act_rate,
            page_index=page_index,
            n_table_rows=n_table_rows,
        )
    elif y_pred == "any":
        return _rank_features_by_cm_value(
            sae_id=sae_id,
            cur=cur,
            key="label_pcts",
            label_index=int(y_true),
            is_descending=is_descending,
            min_act_rate=min_act_rate,
            page_index=page_index,
            n_table_rows=n_table_rows,
        )
    elif y_pred == "different":
        return _rank_features_by_cm_value(
            sae_id=sae_id,
            cur=cur,
            key="false_neg_pcts",
            label_index=int(y_true),
            is_descending=is_descending,
            min_act_rate=min_act_rate,
            page_index=page_index,
            n_table_rows=n_table_rows,
        )
    elif y_true == "any":
        return _rank_features_by_cm_value(
            sae_id=sae_id,
            cur=cur,
            key="pred_label_pcts",
            label_index=int(y_pred),
            is_descending=is_descending,
            min_act_rate=min_act_rate,
            page_index=page_index,
            n_table_rows=n_table_rows,
        )
    elif y_true == "different":
        return _rank_features_by_cm_value(
            sae_id=sae_id,
            cur=cur,
            key="false_pos_pcts",
            label_index=int(y_pred),
            is_descending=is_descending,
            min_act_rate=min_act_rate,
            page_index=page_index,
            n_table_rows=n_table_rows,
        )
    else:
        return _rank_features_by_cm_cell(
            sae_id=sae_id,
            cur=cur,
            true_label_index=int(y_true),
            pred_label_index=int(y_pred),
            is_descending=is_descending,
            min_act_rate=min_act_rate,
            page_index=page_index,
            n_table_rows=n_table_rows,
            n_labels=n_labels,
        )


def _rank_features_by_overall_error_pct(
    sae_id: str,
    cur: sqlite3.Cursor,
    is_descending: bool,
    min_act_rate: float,
    page_index: int,
    n_table_rows: int,
) -> sqlite3.Cursor:
    order = "DESC" if is_descending else "ASC"

    return cur.execute(
        f"""
        SELECT *
        FROM feature
        WHERE sae_id = :sae_id AND sequence_act_rate > :min_act_rate
        ORDER BY JSON_EXTRACT(cm, '$.error_pct') {order}
        LIMIT :limit
        OFFSET :offset
        """,
        {
            "sae_id": sae_id,
            "min_act_rate": min_act_rate,
            "limit": n_table_rows,
            "offset": n_table_rows * page_index,
        },
    )


def _rank_features_by_cm_value(
    sae_id: str,
    cur: sqlite3.Cursor,
    key: str,
    label_index: int,
    is_descending: bool,
    min_act_rate: float,
    page_index: int,
    n_table_rows: int,
) -> sqlite3.Cursor:
    order = "DESC" if is_descending else "ASC"

    return cur.execute(
        f"""
        SELECT *
        FROM feature
        WHERE sae_id = :sae_id AND sequence_act_rate > :min_act_rate
        ORDER BY JSON_EXTRACT(cm, '$.' || :key || '[' || :label_index || ']') {order}
        LIMIT :limit
        OFFSET :offset
        """,
        {
            "sae_id": sae_id,
            "min_act_rate": min_act_rate,
            "key": key,
            "label_index": label_index,
            "limit": n_table_rows,
            "offset": n_table_rows * page_index,
        },
    )


def _rank_features_by_cm_cell(
    sae_id: str,
    cur: sqlite3.Cursor,
    true_label_index: int,
    pred_label_index: int,
    is_descending: bool,
    min_act_rate: float,
    page_index: int,
    n_table_rows: int,
    n_labels: int,
) -> sqlite3.Cursor:
    index = true_label_index * n_labels + pred_label_index
    order = "DESC" if is_descending else "ASC"

    return cur.execute(
        f"""
        SELECT *
        FROM feature
        WHERE sae_id = :sae_id AND sequence_act_rate > :min_act_rate
        ORDER BY JSON_EXTRACT(cm, '$.cells[' || :index || '].pct') {order}
        LIMIT :limit
        OFFSET :offset
        """,
        {
            "sae_id": sae_id,
            "min_act_rate": min_act_rate,
            "index": index,
            "limit": n_table_rows,
            "offset": n_table_rows * page_index,
        },
    )



"""Configuration for SAE analysis."""

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, Literal

if TYPE_CHECKING:
    pass


@dataclass
class AnalysisConfig:
    """Configuration class for analyzing SAEs."""

    # no default

    # dataset
    labels: list[str]

    # default

    # device
    device: 'Literal["cpu", "mps", "cuda", "xpu", "xla"]' = "cuda"
    # dataset
    token_ids_column: str = "input_ids"
    attn_mask_column: str = "attention_mask"
    label_column: str = "label"
    # batch sizes
    model_batch_size_sequences: int = 32
    model_sequence_length: int = 128
    feature_batch_size: int = 256
    # analysis
    total_analysis_tokens: int = 10_000_000
    total_analysis_sequences: int = field(init=False)
    feature_indices: list[int] = field(default_factory=list)
    n_activation_bins: int = 32
    # inferencing
    hidden_state_index: int = -2
    # ui
    n_example_sequences: int = 10
    n_context_tokens: int = 5
    n_sequence_intervals: int = 8
    extra_token_columns: list[str | tuple[str, Callable[[Any], str]]] = field(
        default_factory=list
    )
    extra_sequence_columns: list[str | tuple[str, Callable[[Any], str]]] = field(
        default_factory=list
    )
    # logging
    show_progress: bool = True

    def __post_init__(self):
        self.total_analysis_sequences = (
            self.total_analysis_tokens // self.model_sequence_length
        )

    @classmethod
    def from_dict(cls, data: dict):
        data_copy = data.copy()
        data_copy.pop("total_analysis_sequences")
        return cls(**data_copy)



"""
This is based on the SAE analysis code from sae_vis and SAEDashboard:
https://github.com/callummcdougall/sae_vis
https://github.com/jbloomAus/SAEDashboard
"""

from collections import defaultdict
from typing import TYPE_CHECKING

import numpy as np
import torch
from scipy import stats

from saefarer.analysis.model_and_dataset import get_confusion_matrix
from saefarer.analysis.types import (
    DisplayToken,
    FeatureData,
    FeatureTokenSequence,
    HistogramData,
    MarginalEffectsData,
    SequenceInterval,
    SequenceIntervalIndices,
)

if TYPE_CHECKING:
    import numpy.typing as npt

    from saefarer.analysis.config import AnalysisConfig
    from saefarer.analysis.types import DatasetInfo, ModelInfo
    from saefarer.protocols import TokenizerProtocol


@torch.inference_mode()
def get_feature_data(
    feature_id: int,
    sae_id: str,
    dataset_info: "DatasetInfo",
    model_info: "ModelInfo",
    token_acts: torch.Tensor,
    positive_token_acts_mask: torch.Tensor,
    tokenizer: "TokenizerProtocol",
    ds: dict[str, torch.Tensor],
    cfg: "AnalysisConfig",
    rng: np.random.Generator,
) -> FeatureData:
    # Token activations
    positive_token_acts = token_acts[positive_token_acts_mask]
    token_act_rate = positive_token_acts.numel() / token_acts.numel()
    positive_token_acts_np = positive_token_acts.numpy(force=True)
    token_act_range = (
        positive_token_acts_np.min().item(),
        positive_token_acts_np.max().item(),
    )
    token_acts_histogram = _get_act_histogram(
        positive_token_acts_np, cfg.n_activation_bins, token_act_range
    )

    # Sequence activations
    sequence_acts = token_acts.max(dim=1)[0]
    positive_sequence_acts_mask = sequence_acts > 0
    positive_sequence_acts_mask_cpu = positive_sequence_acts_mask.cpu()
    positive_sequence_acts = sequence_acts[positive_sequence_acts_mask]
    sequence_act_rate = positive_sequence_acts.numel() / sequence_acts.numel()
    positive_sequence_acts_np = positive_sequence_acts.numpy(force=True)
    sequence_act_range = (
        positive_sequence_acts_np.min().item(),
        positive_sequence_acts_np.max().item(),
    )
    sequence_acts_histogram = _get_act_histogram(
        positive_sequence_acts_np, cfg.n_activation_bins, sequence_act_range
    )

    # Example sequences
    sequence_intervals = _get_example_sequences(
        feature_id,
        tokenizer,
        ds,
        token_acts,
        sequence_acts,
        sequence_acts_histogram["thresholds"],
        cfg,
        rng,
    )

    # Marginal effects
    marginal_effects = _get_sequence_level_marginal_effects(
        positive_sequence_acts,
        positive_sequence_acts_mask_cpu,
        sequence_acts_histogram["thresholds"],
        ds,
    )

    # Confusion matrix
    cm = get_confusion_matrix(
        ds[cfg.label_column][positive_sequence_acts_mask_cpu],
        ds["pred_label"][positive_sequence_acts_mask_cpu],
        dataset_info["label_indices"],
    )

    # Additional feature statistics
    mean_pred_label_probs = (
        ds["pred_probs"][positive_sequence_acts_mask_cpu].mean(dim=0).tolist()
    )

    return FeatureData(
        sae_id=sae_id,
        feature_id=feature_id,
        max_act=token_acts.max().item(),
        token_act_rate=token_act_rate,
        token_acts_histogram=token_acts_histogram,
        sequence_act_rate=sequence_act_rate,
        sequence_acts_histogram=sequence_acts_histogram,
        marginal_effects=marginal_effects,
        cm=cm,
        sequence_intervals=sequence_intervals,
        mean_pred_label_probs=mean_pred_label_probs,
    )


@torch.inference_mode()
def _get_example_sequences(
    feature_index: int,
    tokenizer: "TokenizerProtocol",
    ds: dict[str, torch.Tensor],
    token_activations: torch.Tensor,
    sequence_activations: torch.Tensor,
    thresholds: list[float],
    cfg: "AnalysisConfig",
    rng: np.random.Generator,
) -> list[SequenceInterval]:
    interval_indices = _get_interval_indices(sequence_activations, thresholds, cfg, rng)

    sequence_intervals: list[SequenceInterval] = []

    for interval in interval_indices:
        key_seq: list[FeatureTokenSequence] = []

        for seq_i in interval.indices.tolist():
            tok_ids = ds[cfg.token_ids_column][seq_i]
            acts = token_activations[seq_i]
            tok_i = int(torch.argmax(acts).item())

            token_sequence = _get_feature_token_sequence(
                feature_index=feature_index,
                tokenizer=tokenizer,
                input_ids=tok_ids.tolist(),
                activations=acts.tolist(),
                sequence_index=seq_i,
                token_index=tok_i,
                ds=ds,
                cfg=cfg,
            )
            key_seq.append(token_sequence)

        sequence_intervals.append(
            SequenceInterval(
                min_max_act=interval.min_max_act,
                max_max_act=interval.max_max_act,
                sequences=key_seq,
            )
        )

    return sequence_intervals


@torch.inference_mode()
def _get_interval_indices(
    sequence_activations: torch.Tensor,
    thresholds: list[float],
    cfg: "AnalysisConfig",
    rng: np.random.Generator,
) -> list[SequenceIntervalIndices]:
    sequence_indices: list[SequenceIntervalIndices] = []

    top_values, top_indices = torch.topk(
        sequence_activations, k=cfg.n_example_sequences
    )

    positive_mask = top_values > 0
    top_values_positive = top_values[positive_mask]
    top_indices_positive = top_indices[positive_mask]

    sequence_indices.append(
        SequenceIntervalIndices(
            min_max_act=top_values_positive.min().item(),
            max_max_act=top_values_positive.max().item(),
            indices=top_indices_positive,
        )
    )

    bins_per_interval = cfg.n_activation_bins // cfg.n_sequence_intervals
    interval_indices = list(range(0, len(thresholds), bins_per_interval))
    interval_thresholds = [thresholds[i] for i in interval_indices]
    interval_ranges = list(zip(interval_thresholds, interval_thresholds[1:]))

    for i, (interval_min, interval_max) in enumerate(interval_ranges):
        valid_indices = (
            (
                (sequence_activations >= interval_min)
                & (sequence_activations < interval_max)
            )
            .nonzero()
            .flatten()
        )

        if valid_indices.shape[0] > cfg.n_example_sequences:
            # https://stackoverflow.com/a/60564584
            rand_indices = torch.tensor(
                rng.choice(
                    valid_indices.shape[0],
                    cfg.n_example_sequences,
                    replace=False,
                )
            )

            valid_indices = valid_indices[rand_indices]

        sequence_indices.append(
            SequenceIntervalIndices(
                min_max_act=interval_min,
                max_max_act=interval_max,
                indices=valid_indices,
            )
        )

    return sequence_indices


@torch.inference_mode()
def _get_feature_token_sequence(
    feature_index: int,
    tokenizer: "TokenizerProtocol",
    input_ids: list[int],
    activations: list[float],
    sequence_index: int,
    token_index: int,
    ds: dict[str, torch.Tensor],
    cfg: "AnalysisConfig",
) -> FeatureTokenSequence:
    # token metadata columns

    token_extras: dict[str, list[str]] = {}

    for entry in cfg.extra_token_columns:
        if isinstance(entry, str):
            col, formatter = entry, str
        else:
            col, formatter = entry

        values = ds[col][sequence_index].tolist()
        token_extras[col] = [formatter(value) for value in values]

    display_tokens_subset, max_token_index = get_display_tokens(
        tokenizer=tokenizer,
        input_ids=input_ids,
        activations=activations,
        sequence_index=sequence_index,
        token_index=token_index,
        token_extras=token_extras,
        n_context_tokens=cfg.n_context_tokens,
    )

    # sequence metadata

    sequence_extras: dict[str, str] = {}

    for entry in cfg.extra_sequence_columns:
        if isinstance(entry, str):
            col, formatter = entry, str
        else:
            col, formatter = entry

        sequence_extras[col] = formatter(ds[col][sequence_index])

    token_sequence = FeatureTokenSequence(
        feature_index=feature_index,
        sequence_index=sequence_index,
        display_tokens=display_tokens_subset,
        max_token_index=max_token_index,
        label=int(ds["label"][sequence_index]),
        pred_label=int(ds["pred_label"][sequence_index]),
        pred_probs=ds["pred_probs"][sequence_index].tolist(),
        extras=sequence_extras,
    )

    return token_sequence


@torch.inference_mode()
def get_display_tokens(
    tokenizer: "TokenizerProtocol",
    input_ids: list[int],
    activations: list[float],
    sequence_index: int,
    token_index: int,
    token_extras: dict[str, list[str]],
    n_context_tokens: int,
) -> tuple[list[DisplayToken], int]:
    display_tokens: list[DisplayToken] = []

    seq = tokenizer.decode(input_ids)

    token_id_group = []
    activations_group = []
    extras_group = defaultdict(list)

    super_tokens = []

    max_super_token_index = -1

    for i in range(len(input_ids)):
        token_id_group.append(input_ids[i])
        activations_group.append(activations[i])
        for k, v in token_extras.items():
            extras_group[k].append(v[i])

        super_token = tokenizer.decode(token_id_group)

        if i == token_index:
            max_super_token_index = len(display_tokens)

        if seq.startswith("".join(super_tokens) + super_token):
            display_token = DisplayToken(
                display=super_token,
                token_ids=token_id_group,
                acts=activations_group,
                max_act=max(activations_group),
                extras=extras_group,
                # if it's a special token, then it should be the only
                # one in the super token. TODO: make sure this is true
                is_padding=input_ids[i] == tokenizer.pad_token_id,
            )

            display_tokens.append(display_token)
            super_tokens.append(super_token)

            token_id_group = []
            activations_group = []
            extras_group = defaultdict(list)

    assert max_super_token_index != -1

    if token_id_group:
        # this could happen if the original text is longer than the max length
        # for the tokenizer and a multi-token character like an emoji gets split
        print(
            f"Problem decoding {len(token_id_group)} tokens for instance {sequence_index}"
        )

        for token_id, act in zip(token_id_group, activations_group):
            super_token = tokenizer.decode(token_id)
            display_token = DisplayToken(
                display=super_token,
                token_ids=[token_id],
                acts=[act],
                max_act=act,
                extras={},
                is_padding=token_id == tokenizer.pad_token_id,
            )
            display_tokens.append(display_token)

    # take a subset of the tokens

    if n_context_tokens >= 0:
        min_index = max(0, max_super_token_index - n_context_tokens)
        max_index = min(
            len(display_tokens) - 1, max_super_token_index + n_context_tokens
        )

        display_tokens_subset = display_tokens[min_index:max_index]
        max_token_index = max_super_token_index - min_index
    else:
        display_tokens_subset = display_tokens
        max_token_index = max_super_token_index

    return display_tokens_subset, max_token_index


@torch.inference_mode()
def _get_act_histogram(
    acts: "npt.NDArray[np.float64]",
    num_bins: int,
    acts_range: tuple[float, float] | None = None,
) -> HistogramData:
    hist_range = acts_range if acts_range is not None else (acts.min(), acts.max())
    hist, bin_edges = np.histogram(acts, bins=num_bins, range=hist_range)
    return HistogramData(counts=hist.tolist(), thresholds=bin_edges.tolist())


@torch.inference_mode()
def _get_sequence_level_marginal_effects(
    positive_acts: torch.Tensor,
    positive_acts_mask_cpu: torch.Tensor,
    bin_edges: list[float],
    ds: dict[str, torch.Tensor],
) -> MarginalEffectsData:
    pred_probs = ds["pred_probs"][positive_acts_mask_cpu]

    positive_acts_np = positive_acts.numpy(force=True)

    probabilities = []

    for i in range(pred_probs.shape[1]):
        statistic, _, _ = stats.binned_statistic(
            positive_acts_np,
            pred_probs[:, i],
            statistic="mean",
            bins=bin_edges,  # type: ignore
        )

        filled = np.nan_to_num(statistic, nan=-1).tolist()
        probabilities.append(filled)

    non_act_pred_probs = (
        ds["pred_probs"][~positive_acts_mask_cpu].mean(dim=0).nan_to_num(-1).tolist()
    )

    return MarginalEffectsData(
        probs=probabilities, thresholds=bin_edges, non_act_probs=non_act_pred_probs
    )



from typing import TYPE_CHECKING

import torch
import torch.nn.functional as F

from saefarer.analysis.feature_analysis import get_display_tokens
from saefarer.analysis.types import FeatureTokenSequence

if TYPE_CHECKING:
    from transformers import PreTrainedModel

    from saefarer import sae
    from saefarer.analysis import AnalysisConfig
    from saefarer.analysis.types import InferenceInput
    from saefarer.protocols import TokenizerProtocol


def inference(
    inference_input: "InferenceInput",
    model: "PreTrainedModel",
    tokenizer: "TokenizerProtocol",
    sae: "sae.SAE",
    cfg: "AnalysisConfig",
) -> "FeatureTokenSequence":
    tokenize_results = tokenizer.encode(
        inference_input["sequence"], max_length=cfg.model_sequence_length
    )

    input_ids = tokenize_results["token_ids"]
    token_ids = torch.tensor(input_ids, device=sae.device).unsqueeze(0)

    attention_mask = torch.tensor(
        tokenize_results["attention_mask"], device=sae.device
    ).unsqueeze(0)

    model_results = model(
        token_ids,
        attention_mask=attention_mask,
        output_hidden_states=True,
    )
    pred_probs = F.softmax(model_results.logits, dim=1).flatten()
    pred_label: int = pred_probs.argmax().item()  # type: ignore

    model_acts = model_results.hidden_states[cfg.hidden_state_index]
    sae_acts, _ = sae.encode(model_acts)
    feature_acts = sae_acts[0, :, inference_input["feature_index"]]

    display_tokens, max_token_index = get_display_tokens(
        tokenizer=tokenizer,
        input_ids=input_ids,
        activations=feature_acts.tolist(),
        sequence_index=-1,
        token_index=feature_acts.argmax().item(),  # type: ignore
        token_extras={},
        n_context_tokens=-1,
    )

    return FeatureTokenSequence(
        feature_index=inference_input["feature_index"],
        sequence_index=-1,
        display_tokens=display_tokens,
        max_token_index=max_token_index,
        label=-1,
        pred_label=pred_label,
        pred_probs=pred_probs.tolist(),
        extras={},
    )



"""
This is based on the SAE analysis code from sae_vis and SAEDashboard:
https://github.com/callummcdougall/sae_vis
https://github.com/jbloomAus/SAEDashboard
"""

import dataclasses
import datetime
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import torch
import umap
from datasets import (
    Dataset,
    IterableDataset,
)
from torch.utils.data import DataLoader
from tqdm import tqdm

import saefarer.analysis.database as db
from saefarer import sae
from saefarer.analysis.feature_analysis import get_feature_data
from saefarer.analysis.model_and_dataset import (
    get_dataset_info,
    get_dataset_with_predictions,
    get_model_info,
)
from saefarer.analysis.types import (
    FeatureProjection,
    HistogramData,
    SAEData,
)

if TYPE_CHECKING:
    from os import PathLike

    from transformers import PreTrainedModel

    from saefarer.analysis.config import AnalysisConfig
    from saefarer.protocols import TokenizerProtocol


@torch.inference_mode()
def analyze(
    cfg: "AnalysisConfig",
    model: "PreTrainedModel",
    dataset: Dataset | IterableDataset | DataLoader,
    sae: "sae.SAE",
    tokenizer: "TokenizerProtocol",
    output_path: "str | PathLike",
):
    output_path = Path(output_path)

    if output_path.exists():
        raise OSError(f"{output_path} already exists")

    con, cur = db.create_database(output_path)

    model.to(cfg.device)  # type: ignore

    rng = np.random.default_rng()

    ds = get_dataset_with_predictions(model, dataset, cfg)
    dataset_info = get_dataset_info(cfg)
    model_info = get_model_info(ds, dataset_info, cfg)

    db.insert_misc("analysis_cfg", dataclasses.asdict(cfg), con, cur)
    db.insert_misc("dataset_info", dataset_info, con, cur)
    db.insert_misc("model_info", model_info, con, cur)

    # this is in preparation of supporting multiple SAEs
    sae_id = "default"

    feature_indices = cfg.feature_indices or list(range(sae.cfg.d_sae))

    dead_feature_ids, alive_feature_ids = _get_dead_alive_features(sae, feature_indices)
    n_alive_features = len(alive_feature_ids)
    n_dead_features = len(dead_feature_ids)

    feature_batches = [
        alive_feature_ids[i : i + cfg.feature_batch_size]
        for i in range(0, n_alive_features, cfg.feature_batch_size)
    ]

    print("Beginning to analyze features", datetime.datetime.now())

    progress_bar = tqdm(
        total=n_alive_features,
        desc="Calculating feature data",
        disable=not cfg.show_progress,
    )

    non_activating_feature_ids = []
    token_act_rates = []
    sequence_act_rates = []

    sae_activations = torch.zeros(
        ds[cfg.token_ids_column].shape + (cfg.feature_batch_size,),
        device=cfg.device,
        dtype=sae.dtype,
    )

    for features in feature_batches:
        _fill_sae_activations_buffer(sae_activations, features, sae, model, ds, cfg)

        for i, feature in enumerate(features):
            feature_activations = sae_activations[..., i]

            positive_activation_mask = feature_activations > 0

            if positive_activation_mask.sum() == 0:
                non_activating_feature_ids.append(feature)
            else:
                feature_data = get_feature_data(
                    feature,
                    sae_id,
                    dataset_info,
                    model_info,
                    feature_activations,
                    positive_activation_mask,
                    tokenizer,
                    ds,
                    cfg,
                    rng,
                )

                token_act_rates.append(feature_data["token_act_rate"])
                sequence_act_rates.append(feature_data["sequence_act_rate"])

                db.insert_feature(feature_data, con, cur)

            progress_bar.update()

        sae_activations.zero_()

    progress_bar.close()

    print("Finished analyzing features", datetime.datetime.now())

    if non_activating_feature_ids:
        alive_feature_ids = list(
            set(alive_feature_ids) - set(non_activating_feature_ids)
        )
        n_alive_features = len(alive_feature_ids)

    n_non_activating_features = len(non_activating_feature_ids)

    token_act_rate_histogram = _get_activation_rate_histogram(token_act_rates)
    sequence_act_rate_histogram = _get_activation_rate_histogram(sequence_act_rates)

    feature_projection = _get_feature_projection(sae, alive_feature_ids)

    sae_data = SAEData(
        sae_id=sae_id,
        n_total_features=len(feature_indices),
        n_alive_features=n_alive_features,
        n_dead_features=n_dead_features,
        n_non_activating_features=n_non_activating_features,
        alive_feature_ids=alive_feature_ids,
        token_act_rate_histogram=token_act_rate_histogram,
        sequence_act_rate_histogram=sequence_act_rate_histogram,
        feature_projection=feature_projection,
    )

    db.insert_sae(sae_data, con, cur)


@torch.inference_mode()
def _fill_sae_activations_buffer(
    sae_activations: torch.Tensor,
    feature_indices: list[int],
    sae: "sae.SAE",
    model: "PreTrainedModel",
    ds: dict[str, torch.Tensor],
    cfg: "AnalysisConfig",
):
    tokens = ds[cfg.token_ids_column]
    attn_masks = ds[cfg.attn_mask_column]

    offset = 0

    token_batches = tokens.split(cfg.model_batch_size_sequences)
    attn_mask_batches = attn_masks.split(cfg.model_batch_size_sequences)
    # TODO: consider caching these activations so that they don't have to be
    # re-computed for each batch of features.
    for token_batch, attn_mask_batch in zip(token_batches, attn_mask_batches):
        token_batch = token_batch.to(cfg.device)
        attn_mask_batch = attn_mask_batch.to(cfg.device)
        batch_model_output = model(
            token_batch,
            attention_mask=attn_mask_batch,
            output_hidden_states=True,
        )
        batch_model_acts = batch_model_output.hidden_states[cfg.hidden_state_index]
        batch_sae_acts, _ = sae.encode(batch_model_acts)

        start = offset
        offset += batch_sae_acts.shape[0]
        end = offset

        sae_activations[start:end, :, :] = batch_sae_acts[..., feature_indices]


@torch.inference_mode()
def _get_activation_rate_histogram(activation_rates: list[float]) -> HistogramData:
    log_rates = np.log10(activation_rates)
    counts, thresholds = np.histogram(log_rates, "fd")
    return HistogramData(counts=counts.tolist(), thresholds=thresholds.tolist())


@torch.inference_mode()
def _get_dead_alive_features(
    sae: "sae.SAE", feature_indices: list[int]
) -> tuple[list[int], list[int]]:
    dead_mask = sae.get_dead_neuron_mask()

    dead_features = torch.nonzero(dead_mask, as_tuple=True)[0].tolist()
    alive_features = torch.nonzero(~dead_mask, as_tuple=True)[0].tolist()

    user_indices = set(feature_indices)

    dead_features = [i for i in dead_features if i in user_indices]
    alive_features = [i for i in alive_features if i in user_indices]

    return dead_features, alive_features


@torch.inference_mode()
def _get_feature_projection(
    sae: "sae.SAE", feature_ids: list[int]
) -> FeatureProjection:
    n_features = len(feature_ids)

    # UMAP doesn't work with <= 2 datapoints, so in these cases we will
    # return a fake projection.
    if n_features <= 2:
        pos = [float(x) for x in range(n_features)]
        return FeatureProjection(feature_ids=feature_ids, xs=pos, ys=pos)

    weights: np.ndarray = sae.W_dec.numpy(force=True)

    n_neighbors = min(len(feature_ids) - 1, 15)

    reducer = umap.UMAP(n_neighbors=n_neighbors, n_components=2)
    weights_embedded: np.ndarray = reducer.fit_transform(weights[feature_ids])  # type: ignore

    x: list[float] = weights_embedded[:, 0].tolist()
    y: list[float] = weights_embedded[:, 1].tolist()

    return FeatureProjection(feature_ids=feature_ids, xs=x, ys=y)
